local WorldMapInfo = require("src.app.model.WorldMap.WorldMapInfo")
local GetMapNodeListDataCommand = require("src.app.communication.worldMap.GetMapNodeListDataCommand")
local GetMapNodeBlockDataCommand = require("src.app.communication.worldMap.GetMapNodeBlockDataCommand")

--- 更新间隔(当前时间-节点上次更新的信息的秒数大于此值,则加入更新请求列表)
local UPDATE_INTERVAL_SEC = 30 * 1000

local T = Class.WorldMapMgr(LuaEventDispatcher2)

--- 节点信息被更新
T.EVT_TYPE_NODE_INFO_UPDATED = "evt_type_node_info_updated"

--- 地图网格的行数
T.GRID_ROW_NUM = 600

--- 地图风格的列数
T.GRID_COL_NUM = 600

function T:def()
	LuaEventDispatcher2.def(self)

	--- 竞技场信息
	-- _info

	--- 更新时间字典(存储各节点上一次更新的时间)
	-- _updateTimeDict

	--- 待更新列表
	-- _waitUpdateList

	--- 待更新字典
	-- _waitUpdateDict

	--- 待更新计时器
	-- _waitUpdateDelay
end

function T:ctor()
	LuaEventDispatcher2.ctor(self)
end

function T:init()
	self._info = WorldMapInfo()

	self._updateTimeDict = { }
	self._waitUpdateList = { }
	self._waitUpdateDict = { }
end

--- 设置地图节点信息
function T:setNodeData(x, y, data)
	self._info:setNodeData(x, y, data)

	self:luaDispatchEvent(self.EVT_TYPE_NODE_INFO_UPDATED, x, y, data)
end

--- 获取地图节点信息
function T:getNodeData(x, y)
	local lastUpdateSec = self._updateTimeDict[self:_generateKey(x, y)] or 0
	if TimeUtils:getCurrentMillis() - lastUpdateSec > UPDATE_INTERVAL_SEC then
		self:_addWaitUpdateList(x, y)
	end

	return self._info:getNodeData(x, y)
end

----- 获取地图节点的边界数据
--function T:getNodeBoundaryVales(x,y)
--	return self._info:getNodeBoundaryVales(x,y)
--end
--
----- 添加地图区域
--function T:addArea(sx,sy,ex,ey)
--	local l=self._info:addArea(sx,sy,ex,ey)
--	for _,v in pairs(l) do
--		local x,y=v.x,v.y
--
--		--		local val=self:getNodeBoundaryVales(x,y)
--		--		print("地图区域:",x,y,val[1],val[2],val[3],val[4])
--
--		self:luaDispatchEvent(self.EVT_TYPE_NODE_INFO_UPDATED, x, y, self:getNodeData(x,y))
--	end
--end
----- 移除地图区域
--function T:rmvArea(sx,sy,ex,ey)
--	local l=self._info:rmvArea(sx,sy,ex,ey)
--	for _,v in pairs(l) do
--		local x,y=v.x,v.y
--		self:luaDispatchEvent(self.EVT_TYPE_NODE_INFO_UPDATED, x, y, self:getNodeData(x,y))
--	end
--end

--- 强制更新指定坐标列表中的地图节点信息
function T:updateNodesDataForced(nodeCoordinateList)
	for _, v in pairs(nodeCoordinateList) do
		self:_addWaitUpdateList(v.x, v.y)
	end
end

--- 设置地图节点的保护罩状态
function T:setNodeProtectionState(x, y, isProtectioning, playerId)
	local nInfo = self._info:getNodeData(x, y)
	if nInfo then
		if nInfo.playerId == playerId then
			nInfo.isProtection = isProtectioning

			self:luaDispatchEvent(self.EVT_TYPE_NODE_INFO_UPDATED, x, y, nInfo)
		else
			self:_addWaitUpdateList(x, y)
		end
	end
end

--- 加入待更新列表
function T:_addWaitUpdateList(x, y)
	local key = self:_generateKey(x, y)

	-- 检查是否已在待更新列表
	local d = self._waitUpdateDict
	if d[key] == 1 then
		return
	end

	-- 加入待更新列表
	local l = self._waitUpdateList
	l[#l + 1] = { x = x, y = y, key = key }

	-- 延迟1帧后请求数据
	if not self._waitUpdateDelay then
		self._waitUpdateDelay = Cocos2dxUtils:timeOut( function()
			self._waitUpdateDelay = nil

			self:_requestMapNodeInfo()

			self._waitUpdateList = { }
		end , 0)
	end
end

--- 请求地图节点信息
function T:_requestMapNodeInfo()
	local l = self._waitUpdateList

	local count = #l
	if count == 0 then
		return
	end

	--	print("请求的节点数目:", count)

	local callback = function(state, nodeInfoDict)
		local wd = self._waitUpdateDict
		if state == ReturnCode.SUCCESS then
			local t = TimeUtils:getCurrentMillis()
			local ud = self._updateTimeDict
			for _, v in pairs(l) do
				local key = v.key

				ud[key] = t
				wd[key] = nil

				self:setNodeData(v.x, v.y, nodeInfoDict[key])
			end
		else
			for _, v in pairs(l) do
				wd[v.key] = nil
			end
		end
	end

	if count < 40 then
		GetMapNodeListDataCommand():exec(l, callback)
	else
		local sx, sy, ex, ey = 1000, 1000, 0, 0
		for _, v in pairs(l) do
			local x, y = v.x, v.y

			if x < sx then
				sx = x
			elseif x > ex then
				ex = x
			end
			if y < sy then
				sy = y
			elseif y > ey then
				ey = y
			end
		end

		-- 最小坐标到最大坐标构成的区域内的节点个数的80% > 应请求量
		if (ex - sx + 1) *(ey - sy + 1) * 0.8 <= count then
			GetMapNodeBlockDataCommand():exec(sx, sy, ex, ey, callback)
		elseif count < 256 then
			GetMapNodeListDataCommand():exec(l, callback)
		else
			local dl, n = { }, 1
			for _, v in pairs(l) do
				dl[n] = v

				n = n + 1
				if n >= 256 then
					GetMapNodeListDataCommand():exec(dl, callback)

					dl, n = { }, 1
				end
			end
		end
	end
end

--- 生成KEY
function T:_generateKey(x, y)
	return x .. "," .. y
end

return T()
