local AbstractGroundGrid=require("src.frameworks.cocos.ui.ground.AbstractGroundGrid")
local MapGridLogicHandler=require("src.app.uiScript.components.worldMap.MapGridLogicHandler")

local MapCell=require("src.app.uiScript.components.worldMap.MapCell")
local MapNode=require("src.app.uiScript.components.worldMap.MapNode")
local MapNodeType=require("src.app.model.WorldMap.MapNodeType")

local MapNodePlayer=require("src.app.uiScript.components.worldMap.MapNodePlayer")
local MapNodeResLoc=require("src.app.uiScript.components.worldMap.MapNodeResLoc")
local MapNodeSelf=require("src.app.uiScript.components.worldMap.MapNodeSelf")
local MapNodeBorder=require("src.app.uiScript.components.worldMap.MapNodeBorder")

local MapMarchRoute=require("src.app.uiScript.components.worldMap.mapMarchRoute.MapMarchRoute")

--- 点击的事件的触发的最大距离
local TOUCH_MAX_DIS = 16

--- 地图网格
local T=Class.MapGrid(
	function()
		return cc.Node:create()
	end,
	AbstractUI,
	AbstractGroundGrid,
	MapGridLogicHandler
)

--- 网格尺寸-宽
T.CELL_WIDTH=130

--- 网格尺寸-高
T.CELL_HEIGHT=130

--- 节点类型-玩家
T.NODE_TYPE_PLAYER=1

--- 节点类型-资源点
T.NODE_TYPE_RES_LOC=2

--- 节点类型-自己
T.NODE_TYPE_SELF=3

--- 节点类型-边框
T.NODE_TYPE_BORDER=4

function T:def()
	AbstractUI.def(self)
	AbstractGroundGrid.def(self)
	MapGridLogicHandler.def(self)

	--- 网格行数
	self.cellRowNum=0

	--- 网格列数
	self.cellColNum=0

	--- 地图节点对象池
	self.nodePool = {
		[self.NODE_TYPE_PLAYER]=Stack(),
		[self.NODE_TYPE_RES_LOC]=Stack(),
		[self.NODE_TYPE_SELF]=Stack(),
		[self.NODE_TYPE_BORDER]=Stack(),
	}

	--- 正在显示的单元格字典
	self.showingCellDict={}

	--- 点击开始位置
	-- _touchSPt

	--- 行军路线
	-- _uiMapMarchRoute
end

--- 构造函数
-- @param visibleRect 可见区域
-- @param cellSize 单元格尺寸
function T:ctor(visibleRect,cellRowNum,cellColNum)
	AbstractUI.ctor(self)
	MapGridLogicHandler.ctor(self)

	self.cellRowNum=cellRowNum
	self.cellColNum=cellColNum

	AbstractGroundGrid.ctor(self,visibleRect,{width=self.CELL_WIDTH,height=self.CELL_HEIGHT})

	--侦听地图节点信息被改变
	WorldMapMgr:luaAddEventListener(WorldMapMgr.EVT_TYPE_NODE_INFO_UPDATED,self,self._handleNodeUpdate)

	self:_addTouchEvtListener()

	--行军路线
	local uiMapMarchRoute=MapMarchRoute(self)
	self:addChild(uiMapMarchRoute,2)
	self._uiMapMarchRoute=uiMapMarchRoute
end

function T:cleanUp()
	AbstractUI.cleanUp(self)

	WorldMapMgr:luaRemoveEventListener(WorldMapMgr.EVT_TYPE_NODE_INFO_UPDATED,self,self._handleNodeUpdate)

	--释放地图节点对象池中的对象
	print("释放地图节点对象池中的对象")
	for _,stack in pairs(self.nodePool) do
		while not stack:isEmpty() do
			stack:pop():release()
		end
	end
end

--- 更新正在显示的单元格
function T:updateShowingCelles()
	for _,cell in pairs(self.showingCellDict) do
		local x,y=self:coordinateGrid2Map(cell.x,cell.y)
		if x and y then
			WorldMapMgr:getNodeData(x,y)
		end
	end
end

--- 创建单元格
-- @param x
-- @param y
-- @return cell
function T:_createCell(x,y)
	return MapCell(x,y)
end

--- 显示单元格
-- @param cell
-- @return cell
function T:_showCell(cell)
	cell.isShow=true

	local x,y=self:coordinateGrid2Map(cell.x,cell.y)
	if x and y then
		self._uiMapMarchRoute:showCell(cell.x,cell.y)

		local nodeData=WorldMapMgr:getNodeData(x,y)
		if nodeData then
			self:_getShowNode(cell,nodeData)
		else
			self:_putShowNode(cell)
		end

		self.showingCellDict[x..","..y]=cell
	else
		self:_getShowNode(cell,nil)
	end

	return cell
end

--- 隐藏单元格
-- @param cell
-- @return cell
function T:_hideCell(cell)
	cell.isShow=false

	self:_putShowNode(cell)

	local x,y=self:coordinateGrid2Map(cell.x,cell.y)
	if x and y then
		self._uiMapMarchRoute:hideCell(cell.x,cell.y)

		self.showingCellDict[x..","..y]=nil
	end

	return cell
end

--- 获取地图节点
function T:_getShowNode(cell,nodeData)
	local uiNode = cell.uiNode

	local nType
	if nodeData then
		local type=nodeData.type
		if type == MapNodeType.PLAYER then
			if nodeData.playerId == UserInfo.id then
				nType=self.NODE_TYPE_SELF
			else
				nType=self.NODE_TYPE_PLAYER
			end
		elseif type == MapNodeType.RES_LOC then
			nType=self.NODE_TYPE_RES_LOC
		end
	else
		nType=self.NODE_TYPE_BORDER
	end

	if uiNode and uiNode.nodeType ~= nType then
		uiNode:removeFromParent(false)
		self.nodePool[nType]:push(uiNode)

		uiNode=nil
	end

	if not uiNode then
		uiNode=self.nodePool[nType]:pop()
		if not uiNode then
			local w,h=self.CELL_WIDTH,self.CELL_HEIGHT
			if nType == self.NODE_TYPE_SELF then
				uiNode=MapNodeSelf(w,h)
			elseif nType == self.NODE_TYPE_PLAYER then
				uiNode=MapNodePlayer(w,h)
			elseif nType == self.NODE_TYPE_RES_LOC then
				uiNode=MapNodeResLoc(w,h)
			elseif nType == self.NODE_TYPE_BORDER then
				uiNode=MapNodeBorder(w,h)
			end

			uiNode.nodeType=nType
			uiNode:retain()
		end
		
		if not uiNode then
			print("uiNode为什么是 nil 呢?这很奇怪啊,",toString(uiNode),tostring(nType))
			return
		end

		self:addChild(uiNode)
		uiNode:setPosition(self:coordinateGrid2Device(cell.x,cell.y))

		if nType ~= self.NODE_TYPE_BORDER then
			uiNode:setOpacity(0)
			uiNode:runAction(
				cc.FadeIn:create(0.5)
			)
		end
	end

	uiNode:setData(nodeData)

	cell.uiNode=uiNode
	return uiNode
end

--- 存储地图节点
function T:_putShowNode(cell)
	local uiNode = cell.uiNode
	if uiNode then
		cell.uiNode=nil

		uiNode:reset()
		uiNode:removeFromParent(false)
		self.nodePool[uiNode.nodeType]:push(uiNode)
	end
end

--- 处理节点被更新
function T:_handleNodeUpdate(x,y,nodeData)
	local cell=self.showingCellDict[x..","..y]
	if not cell then
		return
	end

	if nodeData then
		self:_getShowNode(cell,nodeData)
	else
		self:_putShowNode(cell)
	end
end

--- 获取正在显示的指定单元格
function T:getCell(x,y)
	return self.showingCellDict[x..","..y]
end

--交互事件处理相关>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
--- 添加touch事件侦听
function T:_addTouchEvtListener()
	local listener = cc.EventListenerTouchOneByOne:create()
	listener:setSwallowTouches(false)

	---开始拖动
	local onTouchBegan = function(t, e)
		self._touchSPt=t:getLocation()

		return true
	end

	--- 拖动结束
	local onTouchEnded = function(t, e)
		local s,e=self._touchSPt,t:getLocation()
		if MathUtil.pointToPointDistance(s.x,s.y,e.x,e.y) <= TOUCH_MAX_DIS then
			local lPt=self:convertToNodeSpace(e)

			local x,y=self:coordinateDevice2Map(lPt.x,lPt.y)
			if not x then
				return
			end

			self:handleTouchEvent(x,y)
		end
	end

	listener:registerScriptHandler(onTouchBegan, cc.Handler.EVENT_TOUCH_BEGAN)
	listener:registerScriptHandler(onTouchEnded, cc.Handler.EVENT_TOUCH_ENDED)
	self:getEventDispatcher():addEventListenerWithSceneGraphPriority(listener, self)
end
--交互事件处理相关<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

--坐标转化与判断>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
--世界坐标原点为左上角,程序坐标原点为左下角
--世界坐标原点为(1,1),程序坐标原点为(0,0)

--- 网格坐标转化为地图坐标
-- @param grdX
-- @param grdY
function T:coordinateGrid2Map(grdX,grdY)
	if self:isInMapArea(grdX,grdY) then
		return grdX+1,self.cellRowNum - grdY
	end
	return nil,nil
end

--- 地图坐标转化为网格坐标
-- @param mapX
-- @param mapY
function T:coordinateMap2Grid(mapX,mapY)
	return mapX-1,self.cellRowNum - mapY
end

--- 设备坐标转化为网格坐标
-- @param dvcX
-- @param dvcY
function T:coordinateDevice2Grid(dvcX,dvcY)
	return math.floor(dvcX / self.CELL_WIDTH),math.floor(dvcY / self.CELL_HEIGHT)
end

--- 网格坐标转化为设备坐标
-- @param grdX
-- @param grdY
function T:coordinateGrid2Device(grdX,grdY)
	return grdX * self.CELL_WIDTH + self.CELL_WIDTH / 2, grdY * self.CELL_HEIGHT + self.CELL_HEIGHT / 2
end

--- 检查指定的网格坐标是否在地图区域内
-- @param grdX
-- @param grdY
function T:isInMapArea(grdX,grdY)
	if grdX < 0 or grdY < 0 or grdX >= self.cellColNum or grdY >= self.cellRowNum then
		return false
	end

	return true
end
--坐标转化与判断<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

return T
