--- 深拷贝对象
-- @param srcObj 对象
local function deepCopy(srcObj)
	--- 已被拷贝的对象列表(解决循环引用的问题)
	local lookup={}

	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup[object] then
			return lookup[object]
		end

		local newTable={}
		lookup[object]=newTable

		for k,v in pairs(object) do
			newTable[_copy(k)]=_copy(v)
		end

		return setmetatable(newTable,getmetatable(object))
	end

	return _copy(srcObj)
end

--- 深拷贝继承
function extend(super)
	local cls
	if super then
		cls=deepCopy(super)
		cls.super=super
	else
		cls={ctor = function() end}
	end
	cls.__index=cls

	function cls.new(...)
		local ins=deepCopy(cls)
		ins.class=cls

		local funcs={}
		local super=cls
		while super do
			if super.super==nil or super.ctor~=super.super.ctor then
				table.insert(funcs,1,super.ctor)
			end
			super=super.super
		end
		for i=1,#funcs do
			funcs[i](ins,...)
		end
		return ins
	end

	function cls:isClass(classObj)
		local cls=self.class
		while cls do
			if cls==classObj then
				return true
			end
			cls=cls.super
		end
		return false
	end

	function cls:asClass(classObj)
		return self:isClass(classObj) and self or nil
	end
	return cls
end

--- 单继承
-- @param 基类,此参数可以指定为 table 表示基类;也可以指定为 function ,返回一个 cocos2dx 对象作为基类
function extend1(baseTemplate)
	local template={}
	template.__index=template

	local createFunction=nil
	if type(baseTemplate) == "function" then
		createFunction=baseTemplate
		baseTemplate=nil
	end

	if createFunction then
		template.__createFunction__=createFunction
	elseif baseTemplate then
		template.__createFunction__=baseTemplate.__createFunction__
	end

	local function __call(cls,...)
		local ins=nil
		if cls.__createFunction__ then
			ins=cls.__createFunction__()

			local t = tolua.getpeer(ins)
			if not t then
				t = {}
				tolua.setpeer(ins, t)
			end
			setmetatable(t,cls)
		else
			ins=setmetatable({},cls)
		end
		ins.super=baseTemplate

		--成员变量定义函数
		if ins.def then
			ins:def()
		end

		--构造函数
		if ins.ctor then
			ins:ctor(...)
		end

		return ins
	end

	if baseTemplate then
		setmetatable(template,{__index=baseTemplate,__call=__call})
	else
		setmetatable(template,{__call=__call})
	end

	return template
end

--- 多继承
-- @param baseTemplate 唯一基类
function inheritMultiple(baseTemplate)
	local template={}
	template.__index=template

	local createFunction=nil
	if type(baseTemplate) == "function" then
		createFunction=baseTemplate
		baseTemplate=nil
	else
		template.__templates__={baseTemplate}
	end

	if createFunction then
		template.__createFunction__=createFunction
	elseif baseTemplate then
		template.__createFunction__=baseTemplate.__createFunction__
	end

	local function __index(o,k)
		for _,v in pairs(o.__templates__) do
			local val=v[k]
			if val then
				o[k]=val
				return val
			end
		end
		return nil
	end

	local function __call(cls,...)
		local ins=nil
		if cls.__createFunction__ then
			ins=cls.__createFunction__()

			local t = tolua.getpeer(ins)
			if not t then
				t = {}
				tolua.setpeer(ins, t)
			end
			setmetatable(t,cls)
		else
			ins=setmetatable({},cls)
		end
		ins.super=baseTemplate

		--成员变量定义函数
		if ins.def then
			ins:def()
		end

		--构造函数
		if ins.ctor then
			ins:ctor(...)
		end

		return ins
	end
	setmetatable(template,{__index=__index,__call=__call})

	return template
end

--- 为多继承实例扩展继承
-- @param ins 多继承实例
-- @param baseInstance 基类实例
function extInherit(ins,baseInstance)
	table.insert(ins.__templates__,baseInstance)
end
