--- 一些数学方法
local MathUtil={}

--- 计算从一点到另一点的角度
-- @param startPoint 开始点
-- @param endPoint 结束点
-- @return 从开始点到结束点的角度
function MathUtil.angleOfDoudlePoint(startPoint,endPoint)
	return math.atan2(endPoint.y-startPoint.y,endPoint.x-startPoint.x) * 180 / math.pi
end

--- 计算圆上某一点坐标
-- @param radian 角度(弧度值)
-- @param r 半径
function MathUtil.getPointOfCircle(radian,r)
	return {y=r*math.sin(radian),x=r*math.cos(radian)}
end

--- 点到点的距离
-- @param x1
-- @param y1
-- @param x2
-- @param y2
function MathUtil.pointToPointDistance(x1,y1,x2,y2)
	local a,b=x1-x2,y1-y2
	return math.sqrt(math.abs((a*a+b*b)))
end

--- 根据两点获取一般直线方程
-- @param x1
-- @param y1
-- @param x2
-- @param y2
function MathUtil.getLineEquationOf2Point(x1,y1,x2,y2)
	-- 直线方程
	-- 一般式:ax+by+c=0
	-- 两点式:(x-x1)/(x2-x1)=(y-y1)/(y2-y1)
	-- 转换过程:
	-- 1.(x-x1)(y2-y1)=(y-y1)(x2-x1)
	-- 2.x(y2-y1)-(x1y2-x1y1)=y(x2-x1)-(x2y1-x1y1)
	-- 3.x(y2-y1)-(x1y2-x1y1)-y(x2-x1)+(x2y1-x1y1)=0
	-- 4.x(y2-y1)-y(x2-x1)-(x1y2-x1y1)+(x2y1-x1y1)=0
	-- 5.x(y2-y1)-y(x2-x1)-x1y2+x1y1+x2y1-x1y1=0
	-- 6.x(y2-y1)-y(x2-x1)-x1y2+x2y1=0
	-- 7.对照一般式:	a=y2-y1
	--					b=-(x2-x1)		=> b=x1-x2
	--					c=-x1y2+x2y1	=> c=x2y1-x1y2
	local a = y2 - y1
	local b = x1 - x2
	local c = x2*y1 - x1*y2
end

--- 计算点到直线的距离
-- @param x 点的横坐标
-- @param y 点的纵坐标
-- @param a 直线公式(a*x+b*y+c=0)参数a
-- @param b 直线公式(a*x+b*y+c=0)参数b
-- @param c 直线公式(a*x+b*y+c=0)参数c
function MathUtil.pointToLineDistance(x,y,a,b,c)
	return math.abs(a*x+b*y+c)/math.sqrt(a*a+b*b)
end

--- 知道直线上的某一点的y坐标,求x坐标
-- @param y 直线上某一点的y坐标
-- @param a 直线公式(a*x+b*y+c=0)参数a
-- @param b 直线公式(a*x+b*y+c=0)参数b
-- @param c 直线公式(a*x+b*y+c=0)参数c
function MathUtil.xOfLine(y,a,b,c)
	return (-b*y-c)/a --直线的公式:a*x+b*y+c=0,所以:x=(-b*y-c)/a
end

--- 知道直线上的某一点的x坐标,求y坐标
-- @param x 直线上某一点的x坐标
-- @param a 直线公式(a*x+b*y+c=0)参数a
-- @param b 直线公式(a*x+b*y+c=0)参数b
-- @param c 直线公式(a*x+b*y+c=0)参数c
function MathUtil.yOfLine(x,a,b,c)
	return (-a*x-c)/b --直线的公式:a*x+b*y+c=0,所以:y=(-a*x-c)/b
end

--- 判断点是否被包含在多边形内(射线法)
-- @param p 点
-- @param polygon 构成多边形的点列表
function MathUtil.containsPointInPolygon(p,polygon)
	local c=0
	local len=#polygon

	for i=1,len,1 do
		local p1,p2=polygon[i]
		if i == len then
			p2=polygon[1]
		else
			p2=polygon[i+1]
		end

		if p1.y == p2.y --p1p2 与 y=p0.y平行
			or
			p.y < math.min(p1.y,p2.y) --交点在p1p2延长线上
			or
			p.y >= math.max(p1.y,p2.y) --交点在p1p2延长线上
		then

		else
			--交点的 X 坐标
			local x=(p.y-p1.y)*(p2.x-p1.x)/(p2.y-p1.y)+p1.x
			if x > p.x then --向右水平方向发射射线
				c=c+1
			end
		end
	end

	return c % 2 == 1
end

--- 获取组成圆的点数组
-- @param cx 中心点 x 坐标
-- @param cy 中心点 y 坐标
-- @param radius 半径
-- @param step 步长
-- @return point 数组
function MathUtil:getPointsOfCycle(cx,cy,radius,step)
	local pts,r={},0

	local cos,sin,rad=math.cos,math.sin,math.rad
	for i=0,359,step do
		r=rad(i)
		pts[#pts+1]={
			x=cx+radius * cos(r),
			y=cy+radius * sin(r)
		}
	end

	return pts
end

--- 在圆内随机生成一点
-- 即计算圆内随机半径的同心圆上的随机点
function MathUtil:randomInCycle(radius)
	--同心圆半径
	local ir=randomFloat(0,radius)
	--任意角度
	local angle=randomFloat(0,math.pi)*2

	local t=math.sqrt(ir)
	return {
		x=math.sin(angle)*t,
		y=math.cos(angle)*t
	}
end

--- 获取三角形内随机一点
-- 以下为数学原理
-- 已知了三角形三个点a,b,c，那么可以求得两个向量ab, ac。
-- 已经有一个rand(a, b)， 那么使用rand(0, 1)获得两个0～1之间的随机实数x, y
-- 如果x+y>1， 那么令x'=1-x, y'=1-y
-- 如果x+y<=1, 那么令x'=x, y'=y
-- 那么：
-- 目标点 = 点a + x' * 向量ab + y‘ * 向量ac
-- 这么做的好处是一来符合平均分布，二来不用反复试错，一次出结果。
function MathUtil:randomInTriangle(a,b,c)
	local m,n=randomFloat(0,1),randomFloat(0,1)

	if m + n > 1 then
		m,n=1-m,1-n
	end

	return {
		x=a.x+m*(b.x-a.x)+n*(c.x-a.x),
		y=a.y+m*(b.y-a.y)+n*(c.y-a.y)
	}
end

return MathUtil
